use std::collections::HashMap;

use crate::model::{
    log::Log,
    role::Role,
    user::{AdminUser, ManageUserUser, User, UserWithDetails},
};
use futures::future::join_all;
use rocket::{
    form::Form,
    get,
    http::Status,
    post,
    request::{FlashMessage, FromRequest, Outcome},
    response::{Flash, Redirect},
    routes, FromForm, Request, Route, State,
};
use rocket_dyn_templates::{tera::Context, Template};
use sqlx::SqlitePool;

// Custom request guard to extract the Referer header
struct Referer(String);

#[rocket::async_trait]
impl<'r> FromRequest<'r> for Referer {
    type Error = ();

    async fn from_request(request: &'r Request<'_>) -> Outcome<Self, Self::Error> {
        match request.headers().get_one("Referer") {
            Some(referer) => Outcome::Success(Referer(referer.to_string())),
            None => Outcome::Error((Status::BadRequest, ())),
        }
    }
}

#[get("/user")]
async fn index(
    db: &State<SqlitePool>,
    user: ManageUserUser,
    flash: Option<FlashMessage<'_>>,
) -> Template {
    let user_futures: Vec<_> = User::all(db)
        .await
        .into_iter()
        .map(|u| async move { UserWithDetails::from_user(u, db).await })
        .collect();

    let user: User = user.into_inner();
    let allowed_to_edit = ManageUserUser::new(db, user.clone()).await.is_some();

    let users: Vec<UserWithDetails> = join_all(user_futures).await;

    let roles = Role::all(db).await;

    let mut context = Context::new();
    if let Some(msg) = flash {
        context.insert("flash", &msg.into_inner());
    }
    context.insert("allowed_to_edit", &allowed_to_edit);
    context.insert("users", &users);
    context.insert("roles", &roles);
    context.insert("loggedin_user", &UserWithDetails::from_user(user, db).await);

    Template::render("admin/user/index", context.into_json())
}

#[get("/user", rank = 2)]
async fn index_admin(
    db: &State<SqlitePool>,
    user: AdminUser,
    flash: Option<FlashMessage<'_>>,
) -> Template {
    let user_futures: Vec<_> = User::all(db)
        .await
        .into_iter()
        .map(|u| async move { UserWithDetails::from_user(u, db).await })
        .collect();
    let users: Vec<UserWithDetails> = join_all(user_futures).await;

    let user: User = user.user;
    let allowed_to_edit = ManageUserUser::new(db, user.clone()).await.is_some();

    let roles = Role::all(db).await;

    let mut context = Context::new();
    if let Some(msg) = flash {
        context.insert("flash", &msg.into_inner());
    }
    context.insert("allowed_to_edit", &allowed_to_edit);
    context.insert("users", &users);
    context.insert("roles", &roles);
    context.insert("loggedin_user", &UserWithDetails::from_user(user, db).await);

    Template::render("admin/user/index", context.into_json())
}

#[get("/user/<user>/reset-pw")]
async fn resetpw(db: &State<SqlitePool>, admin: ManageUserUser, user: i32) -> Flash<Redirect> {
    let user = User::find_by_id(db, user).await;
    match user {
        Some(user) => {
            Log::create(
                db,
                format!("{} has resetted the pw for {}", admin.user.name, user.name),
            )
            .await;
            user.reset_pw(db).await;
            Flash::success(
                Redirect::to("/admin/user"),
                format!("Passwort von {} zurückgesetzt", user.name),
            )
        }
        None => Flash::error(Redirect::to("/admin/user"), "User does not exist"),
    }
}

#[get("/user/<user>/delete")]
async fn delete(db: &State<SqlitePool>, admin: ManageUserUser, user: i32) -> Flash<Redirect> {
    let user = User::find_by_id(db, user).await;
    Log::create(db, format!("{} deleted user: {user:?}", admin.user.name)).await;
    match user {
        Some(user) => {
            user.delete(db).await;
            Flash::success(
                Redirect::to("/admin/user"),
                format!("Benutzer {} gelöscht", user.name),
            )
        }
        None => Flash::error(Redirect::to("/admin/user"), "User does not exist"),
    }
}

#[derive(FromForm, Debug)]
pub struct UserEditForm {
    pub(crate) id: i32,
    pub(crate) name: String,
    pub(crate) roles: HashMap<String, String>,
}

#[post("/user", data = "<data>", format = "multipart/form-data")]
async fn update(
    db: &State<SqlitePool>,
    data: Form<UserEditForm>,
    admin: ManageUserUser,
) -> Flash<Redirect> {
    let user = User::find_by_id(db, data.id).await;
    Log::create(
        db,
        format!("{} updated user from {user:?} to {data:?}", admin.user.name),
    )
    .await;
    let Some(user) = user else {
        return Flash::error(
            Redirect::to("/admin/user"),
            format!("User with ID {} does not exist!", data.id),
        );
    };

    match user.update(db, data.into_inner()).await {
        Ok(_) => Flash::success(
            Redirect::to("/admin/user"),
            "Mitglied erfolgreich geändert!",
        ),
        Err(e) => Flash::error(Redirect::to("/admin/user"), e),
    }
}

#[derive(FromForm, Debug)]
struct UserAddForm<'r> {
    name: &'r str,
}

#[post("/user/new", data = "<data>")]
async fn create(
    db: &State<SqlitePool>,
    data: Form<UserAddForm<'_>>,
    admin: ManageUserUser,
) -> Flash<Redirect> {
    User::create(db, data.name).await;

    Log::create(
        db,
        format!("{} created new user: {data:?}", admin.user.name),
    )
    .await;

    Flash::success(
        Redirect::to("/admin/user"),
        "Mitglied erfolgreich angelegt!",
    )
}

pub fn routes() -> Vec<Route> {
    routes![index, index_admin, resetpw, update, create, delete]
}
